# Copyright 2020-2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np

# Overflow warnings are expected, ignore them
warnings.simplefilter("ignore", RuntimeWarning)

MAX_CHUNK_LENGTH = 0x1FFF


# MurmurHash3 "Mix4" variant
def RandomHelper_seed(x):
    z = (np.uint64(x) ^ (np.uint64(x) >> np.uint64(33))) * np.uint64(0x62A9D9ED799705F5)
    y = (np.uint64(z) ^ (np.uint64(z) >> np.uint64(28))) * np.uint64(0xCB24D0A5C88C35B3)
    return np.int64(np.uint64(y) >> np.uint64(32))


def RandomHelper_next(state):
    s0 = np.int16(np.uint64(state) & np.uint64(0xFFFF))
    s1 = np.int16(np.uint64(state) >> np.uint64(16) & np.uint64(0xFFFF))
    next = s0
    next += s1
    next = RandomHelper_rotl(next, 9)
    next += s0

    s1 = np.int16(np.uint16(s1) ^ np.uint16(s0))
    s0 = RandomHelper_rotl(s0, 13)
    s0 = np.int16(np.uint16(s0) ^ np.uint16(s1))
    s0 = np.int16(np.uint16(s0) ^ np.uint16(np.uint16(s1) << np.uint16(5)))
    s1 = RandomHelper_rotl(s1, 10)

    result = next
    result <<= 16
    result |= s1
    result <<= 16
    result |= s0
    return result


def RandomHelper_rotl(x, k):
    return np.int16((np.uint32(x) << np.uint32(k)) | (np.uint32(x) >> (32 - np.uint32(k))))


def DeobfuscatorHelper_getString(id, chunks):
    state = RandomHelper_seed(np.array(id).astype(np.uint64) & np.uint64(0xFFFFFFFF))
    state = RandomHelper_next(state)
    low = np.int64((np.array(state).astype(np.uint64) >> np.uint64(32)) & np.uint64(0xFFFF))
    state = RandomHelper_next(state)
    high = np.int64((np.array(state).astype(np.uint64) >> np.uint64(16)) & np.uint64(0xFFFF0000))
    index = np.int32(
        (np.array(id).astype(np.uint64) >> np.uint64(32)) ^ np.uint64(low) ^ np.uint64(high)
    )
    state = DeobfuscatorHelper_getCharAt(index, chunks, state)
    length = np.int32((np.array(state).astype(np.uint64) >> np.uint64(32)) & np.uint64(0xFFFF))

    output = []

    for x in range(length):
        state = DeobfuscatorHelper_getCharAt(index + x + 1, chunks, state)
        output.append(state >> 32 & 0xFFFF)

    return "".join(map(chr, output)).encode("unicode-escape")


def DeobfuscatorHelper_getCharAt(char_index, chunks, state):
    next_state = RandomHelper_next(state)
    chunk = chunks[int(char_index / MAX_CHUNK_LENGTH)]

    return np.int64(
        np.uint64(next_state)
        ^ (
            np.uint64(
                np.frombuffer(
                    chunk[int(char_index % MAX_CHUNK_LENGTH)].encode("utf-16", "surrogatepass")[
                        2:
                    ],
                    dtype=np.uint16,
                )[0]
            )
            << np.uint16(32)
        )
    )



# Copyright 2020-2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List

from .deobfuscator import DeobfuscatorHelper_getString

"""
const-wide v0, -0x123L

invoke-static {v0, v1}, Lorg/lsposed/lsparanoid/Deobfuscator;->getString(J)Ljava/lang/String;

move-result-object v0
"""

"""
const-string v0, "DEOBFUSCATED"
"""

MOVE_RESULT_OBJECT_REGEX = re.compile(r"move-result-object\s+([vp][0-9]+)")
CONST_STRING_REGEX = re.compile(r'const-string\s+[vp][0-9]+,\s+"((?:\\u[0-9a-f]{4})+)"')
CONST_WIDE_REGEX = re.compile(
    r"const-wide(?:/(?:high)?(?:16|32))?\s+([vp][0-9]+),\s+(-?0x[a-f0-9]+)L?"
)
INVOKE_STATIC_REGEX = re.compile(r"invoke-static\s+{([vp][0-9]+),\s+[vp][0-9]+},")
UTF16_REGEX = re.compile(r"^(?:\\u[0-9a-f]{4})+$")
# TODO: Match DEX SimpleName instead of \w
DEOBFUSCATOR_SIGNATURE_REGEX = re.compile(
    r"\.method\s+public\s+static\s+(\w+)\(J\)Ljava/lang/String;"
)


def is_encrypted_string(data: str):
    # Check if it's a UTF-16 string
    if not UTF16_REGEX.match(data):
        return False

    # Minimun count of \\u is 2
    if data.count("\\u") < 2:
        return False

    return True


def extract_obfuscated_strings(data: str):
    result = []

    for line in data.splitlines():
        line = line.strip()

        # Filter out lines
        if not line or not line.startswith("const-string"):
            continue

        m = CONST_STRING_REGEX.search(line)
        if m and is_encrypted_string(m.group(1)):
            result.append(m.group(1))

    return result


def extract_deobfuscator_method(data: str):
    """
    Search for this signature:

    Lorg/lsposed/lsparanoid/Deobfuscator;->getString(J)Ljava/lang/String;
    """

    for line in data.splitlines():
        # Filter out lines
        line = line.strip()
        if not line or not line.startswith(".method"):
            continue

        m = DEOBFUSCATOR_SIGNATURE_REGEX.search(line)
        if m:
            return "{}(J)Ljava/lang/String;".format(m.group(1))

    return None


def extract_identifier_from_invoke(line: str):
    m = INVOKE_STATIC_REGEX.search(line)
    if m:
        return m.group(1)


def deobfuscate_strings(data: str, signature: str, chunks: List[str]):
    result = []

    const_buffer = []

    for line_num, line in enumerate(data.splitlines()):
        # Filter out lines
        line = line.strip()
        if not line:
            continue

        # Clear buffer if end method
        if line == ".end method":
            const_buffer.clear()

        # Add long to buffer
        if line.startswith("const-wide"):
            m = CONST_WIDE_REGEX.search(line)
            if m:
                const_buffer.append((m.group(1), m.group(2)))

        # Check if calling deobfuscator method
        if signature in line:
            m = INVOKE_STATIC_REGEX.search(line)
            if m:
                identifier = m.group(1)
                for x in reversed(const_buffer):
                    if x[0] == identifier:
                        string_id = int(x[1], 16)
                        decoded = (
                            DeobfuscatorHelper_getString(string_id, chunks)
                            .decode()
                            .replace('"', '\\"')
                            .replace("'", "\\'")
                            .replace("\\x", "\\u00")
                        )
                        result.append((line_num, decoded))
                        break

    return result



#!/usr/bin/env python3
# Copyright 2020-2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import logging
import pathlib
import sys
from typing import List

import paranoid_deobfuscator as deobfuscator

REMOVED_COMMENT = (
    "    # Removed with https://github.com/giacomoferretti/paranoid-deobfuscator\n"
)
INSTRUCTION_BUFFER = 20


def search_candidates(target: pathlib.Path):
    candidates = []

    for x in target.rglob("*"):
        if x.is_file() and x.suffix == ".smali":
            with open(x) as f:
                chunks = deobfuscator.extract_obfuscated_strings(f.read())
                if len(chunks) > 0:
                    candidates.append((x, chunks))

    return candidates


def search_deobfuscator(
    target: pathlib.Path, candidate: pathlib.Path, is_apktool: bool
):
    with open(candidate) as f:
        method = deobfuscator.extract_deobfuscator_method(f.read())

        if method:
            clazz = candidate.relative_to(target)
            if is_apktool:
                clazz = clazz.relative_to(*clazz.parts[:1])

            return f"L{'/'.join(clazz.with_suffix('').parts)};->{method}"

    return None


def decode_unicode(chunks: List[str]):
    return [x.encode().decode("unicode-escape") for x in chunks]


def main(args):
    logger = logging.getLogger(__name__)

    # TODO: Alternative for intellisense?
    target: pathlib.Path = args.target
    keep_calls: bool = args.keep_calls

    # Check if target is directory
    if not target.is_dir():
        logger.error("Target must be a directory containing .smali files.")
        sys.exit(1)

    # Check if target is Apktool's output
    is_apktool = (target / "apktool.yml").is_file()
    if is_apktool:
        logger.debug("Target is Apktool output.")

    # Check if specified deobfuscator is correct
    # if not args.deobfuscator.is_file() or not args.deobfuscator.suffix == ".smali":
    #     logger.warning("Specified deobfuscator file is not a .smali file. Fallback to search.")

    # Step 1 - Search for UTF-16 strings (obfuscated chunks)
    logger.debug("Searching for obfuscated strings...")
    candidates = search_candidates(target)

    # Exit if no candidates found
    if len(candidates) == 0:
        logger.error("No candidate found. Is this app really using Paranoid?")
        sys.exit(1)

    logger.debug(f"Found {len(candidates)} candidates.")
    for x in candidates:
        logger.debug(f" - {x[0]} [chunks={len(x[1])}]")

    # Step 2 - Search for deobfuscator class
    logger.debug("Searching for deobfuscator class...")
    getstring_method_signature, chunks = None, None
    for x in candidates:
        getstring_method_signature = search_deobfuscator(target, x[0], is_apktool)
        if getstring_method_signature:
            chunks = x[1]
            break

    # Check
    if not getstring_method_signature or not chunks:
        logger.error(
            "Cannot find Deobfuscator.getString() method. Is this app really using Paranoid?"
        )
        sys.exit(1)

    logger.debug(f"getString method signature: {getstring_method_signature}")

    # 3. Search getString() calls
    logger.info("Searching for getString() calls...")

    # Decode chunks
    chunks = decode_unicode(chunks)

    obfuscated_files = {}
    for x in target.rglob("*"):
        if x.is_file() and x.suffix == ".smali":
            with open(x) as f:
                deobfuscated = deobfuscator.deobfuscate_strings(
                    f.read(), getstring_method_signature, chunks
                )

                if len(deobfuscated) > 0:
                    obfuscated_files[str(x)] = deobfuscated

    # Debug print
    logger.debug(f"Obfuscated files: {len(obfuscated_files)}")
    for k, v in obfuscated_files.items():
        logger.debug(f" - {k}")
        for x in v:
            logger.debug(f"   - {x[0]}: {x[1]}")

    # 4. Deobfuscate
    for file, decoded_strings in obfuscated_files.items():
        logger.info(f"Processing {file}")
        with open(file) as f:
            lines = f.readlines()

        for decoded in decoded_strings:
            line_num, deobfuscated = decoded

            identifier = deobfuscator.extract_identifier_from_invoke(lines[line_num])
            logger.debug(f" - identifier {identifier}")

            # Clear matching const-wide
            for i, x in enumerate(
                reversed(lines[line_num - INSTRUCTION_BUFFER : line_num])
            ):
                x = x.strip()

                # Skip empty lines, comments, labels and directives
                if not x or x.startswith("#") or x.startswith(".") or x.startswith(":"):
                    continue

                if x.startswith("const-wide"):
                    m = deobfuscator.CONST_WIDE_REGEX.match(x)
                    if m and identifier == m.group(1):
                        lines[line_num - i - 1] = REMOVED_COMMENT
                        break

            # Search for move-result-object
            # The next instruction should be move-result-object, because getString() returns a String object
            # If not, it's probably a useless call that was optimized
            for i, x in enumerate(
                lines[line_num + 1 : line_num + 1 + INSTRUCTION_BUFFER]
            ):
                x = x.strip()

                # Skip empty lines, comments, labels and directives
                if not x or x.startswith("#") or x.startswith(".") or x.startswith(":"):
                    continue

                # If next instruction is not move-result-object, break
                if not x.startswith("move-result-object"):
                    logger.debug(
                        f'Could not find move-result-object, probably useless call (string was: "{deobfuscated}")'
                    )
                    break

                # Extract identifier
                m = deobfuscator.MOVE_RESULT_OBJECT_REGEX.match(x)
                if m:
                    lines[
                        line_num + 1 + i
                    ] = f'    const-string {m.group(1)}, "{deobfuscated}"\n'
                break

            # Clear invoke-static call
            if not keep_calls:
                lines[line_num] = REMOVED_COMMENT

        with open(file, "w") as f:
            f.writelines(lines)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Deobfuscate paranoid string encryption."
    )
    parser.add_argument(
        "target",
        help="apktool folder or dex file",
        type=pathlib.Path,
    )

    parser.add_argument(
        "-k",
        "--keep-calls",
        help="keep getString() calls",
        action="store_true",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.WARNING,
    )

    args = parser.parse_args()
    logging.basicConfig(format="%(levelname).1s: %(message)s", level=args.loglevel)

    main(args)
